@@ -25,20 +25,23 @@
 #include "llvm/Analysis/LoopInfo.h"
 #include "llvm/Analysis/ScalarEvolution.h"
 #include "llvm/Analysis/ScalarEvolutionExpander.h"
+#include "llvm/IR/IntrinsicInst.h"
 #include "llvm/Transforms/Utils/BasicBlockUtils.h"
 
 using namespace llvm;
 using namespace polly;
 
 static cl::opt<bool>
-Aligned("enable-polly-aligned", cl::desc("Assumed aligned memory accesses."),
-        cl::Hidden, cl::value_desc("OpenMP code generation enabled if true"),
-        cl::init(false), cl::ZeroOrMore, cl::cat(PollyCategory));
+    Aligned("enable-polly-aligned",
+            cl::desc("Assumed aligned memory accesses."), cl::Hidden,
+            cl::value_desc("OpenMP code generation enabled if true"),
+            cl::init(false), cl::ZeroOrMore, cl::cat(PollyCategory));
 
 static cl::opt<bool, true>
-SCEVCodegenF("polly-codegen-scev", cl::desc("Use SCEV based code generation."),
-             cl::Hidden, cl::location(SCEVCodegen), cl::init(false),
-             cl::ZeroOrMore, cl::cat(PollyCategory));
+    SCEVCodegenF("polly-codegen-scev",
+                 cl::desc("Use SCEV based code generation."), cl::Hidden,
+                 cl::location(SCEVCodegen), cl::init(false), cl::ZeroOrMore,
+                 cl::cat(PollyCategory));
 
 bool polly::SCEVCodegen;
 
@@ -64,7 +67,7 @@ bool polly::canSynthesize(const Instruction *I, const llvm::LoopInfo *LI,
 namespace {
 class IslGenerator {
 public:
-  IslGenerator(IRBuilder<> &Builder, std::vector<Value *> &IVS)
+  IslGenerator(PollyIRBuilder &Builder, std::vector<Value *> &IVS)
       : Builder(Builder), IVS(IVS) {}
   Value *generateIslVal(__isl_take isl_val *Val);
   Value *generateIslAff(__isl_take isl_aff *Aff);
@@ -76,7 +79,7 @@ private:
     class IslGenerator *Generator;
   } IslGenInfo;
 
-  IRBuilder<> &Builder;
+  PollyIRBuilder &Builder;
   std::vector<Value *> &IVS;
   static int mergeIslAffValues(__isl_take isl_set *Set, __isl_take isl_aff *Aff,
                                void *User);
@@ -132,7 +135,7 @@ int IslGenerator::mergeIslAffValues(__isl_take isl_set *Set,
                                     __isl_take isl_aff *Aff, void *User) {
   IslGenInfo *GenInfo = (IslGenInfo *)User;
 
-  assert((GenInfo->Result == NULL) &&
+  assert((GenInfo->Result == nullptr) &&
          "Result is already set. Currently only single isl_aff is supported");
   assert(isl_set_plain_is_universe(Set) &&
          "Code generation failed because the set is not universe");
@@ -145,7 +148,7 @@ int IslGenerator::mergeIslAffValues(__isl_take isl_set *Set,
 
 Value *IslGenerator::generateIslPwAff(__isl_take isl_pw_aff *PwAff) {
   IslGenInfo User;
-  User.Result = NULL;
+  User.Result = nullptr;
   User.Generator = this;
   isl_pw_aff_foreach_piece(PwAff, mergeIslAffValues, &User);
   assert(User.Result && "Code generation for isl_pw_aff failed");
@@ -154,7 +157,7 @@ Value *IslGenerator::generateIslPwAff(__isl_take isl_pw_aff *PwAff) {
   return User.Result;
 }
 
-BlockGenerator::BlockGenerator(IRBuilder<> &B, ScopStmt &Stmt, Pass *P)
+BlockGenerator::BlockGenerator(PollyIRBuilder &B, ScopStmt &Stmt, Pass *P)
     : Builder(B), Statement(Stmt), P(P), SE(P->getAnalysis<ScalarEvolution>()) {
 }
 
@@ -185,7 +188,7 @@ Value *BlockGenerator::lookupAvailableValue(const Value *Old, ValueMapT &BBMap,
   if (Value *New = BBMap.lookup(Old))
     return New;
 
-  return NULL;
+  return nullptr;
 }
 
 Value *BlockGenerator::getNewValue(const Value *Old, ValueMapT &BBMap,
@@ -214,18 +217,21 @@ Value *BlockGenerator::getNewValue(const Value *Old, ValueMapT &BBMap,
   // Now the scalar dependence is neither available nor SCEVCodegenable, this
   // should never happen in the current code generator.
   llvm_unreachable("Unexpected scalar dependence in region!");
-  return NULL;
+  return nullptr;
 }
 
 void BlockGenerator::copyInstScalar(const Instruction *Inst, ValueMapT &BBMap,
                                     ValueMapT &GlobalMap, LoopToScevMapT &LTS) {
+  // We do not generate debug intrinsics as we did not investigate how to
+  // copy them correctly. At the current state, they just crash the code
+  // generation as the meta-data operands are not correctly copied.
+  if (isa<DbgInfoIntrinsic>(Inst))
+    return;
+
   Instruction *NewInst = Inst->clone();
 
   // Replace old operands with the new ones.
-  for (Instruction::const_op_iterator OI = Inst->op_begin(),
-                                      OE = Inst->op_end();
-       OI != OE; ++OI) {
-    Value *OldOperand = *OI;
+  for (Value *OldOperand : Inst->operands()) {
     Value *NewOperand =
         getNewValue(OldOperand, BBMap, GlobalMap, LTS, getLoopForInst(Inst));
 
@@ -249,7 +255,6 @@ void BlockGenerator::copyInstScalar(const Instruction *Inst, ValueMapT &BBMap,
 std::vector<Value *> BlockGenerator::getMemoryAccessIndex(
     __isl_keep isl_map *AccessRelation, Value *BaseAddress, ValueMapT &BBMap,
     ValueMapT &GlobalMap, LoopToScevMapT &LTS, Loop *L) {
-
   assert((isl_map_dim(AccessRelation, isl_dim_out) == 1) &&
          "Only single dimensional access functions supported");
 
@@ -382,12 +387,11 @@ void BlockGenerator::copyBB(ValueMapT &GlobalMap, LoopToScevMapT &LTS) {
 
   ValueMapT BBMap;
 
-  for (BasicBlock::const_iterator II = BB->begin(), IE = BB->end(); II != IE;
-       ++II)
-    copyInstruction(II, BBMap, GlobalMap, LTS);
+  for (Instruction &Inst : *BB)
+    copyInstruction(&Inst, BBMap, GlobalMap, LTS);
 }
 
-VectorBlockGenerator::VectorBlockGenerator(IRBuilder<> &B,
+VectorBlockGenerator::VectorBlockGenerator(PollyIRBuilder &B,
                                            VectorValueMapT &GlobalMaps,
                                            std::vector<LoopToScevMapT> &VLTS,
                                            ScopStmt &Stmt,
@@ -431,12 +435,18 @@ Type *VectorBlockGenerator::getVectorPtrTy(const Value *Val, int Width) {
   return PointerType::getUnqual(VectorType);
 }
 
-Value *VectorBlockGenerator::generateStrideOneLoad(const LoadInst *Load,
-                                                   ValueMapT &BBMap) {
+Value *
+VectorBlockGenerator::generateStrideOneLoad(const LoadInst *Load,
+                                            VectorValueMapT &ScalarMaps,
+                                            bool NegativeStride = false) {
+  unsigned VectorWidth = getVectorWidth();
   const Value *Pointer = Load->getPointerOperand();
-  Type *VectorPtrType = getVectorPtrTy(Pointer, getVectorWidth());
-  Value *NewPointer =
-      getNewValue(Pointer, BBMap, GlobalMaps[0], VLTS[0], getLoopForInst(Load));
+  Type *VectorPtrType = getVectorPtrTy(Pointer, VectorWidth);
+  unsigned Offset = NegativeStride ? VectorWidth - 1 : 0;
+
+  Value *NewPointer = nullptr;
+  NewPointer = getNewValue(Pointer, ScalarMaps[Offset], GlobalMaps[Offset],
+                           VLTS[Offset], getLoopForInst(Load));
   Value *VectorPtr =
       Builder.CreateBitCast(NewPointer, VectorPtrType, "vector_ptr");
   LoadInst *VecLoad =
@@ -444,6 +454,16 @@ Value *VectorBlockGenerator::generateStrideOneLoad(const LoadInst *Load,
   if (!Aligned)
     VecLoad->setAlignment(8);
 
+  if (NegativeStride) {
+    SmallVector<Constant *, 16> Indices;
+    for (int i = VectorWidth - 1; i >= 0; i--)
+      Indices.push_back(ConstantInt::get(Builder.getInt32Ty(), i));
+    Constant *SV = llvm::ConstantVector::get(Indices);
+    Value *RevVecLoad = Builder.CreateShuffleVector(
+        VecLoad, VecLoad, SV, Load->getName() + "_reverse");
+    return RevVecLoad;
+  }
+
   return VecLoad;
 }
 
@@ -504,11 +524,17 @@ void VectorBlockGenerator::generateLoad(const LoadInst *Load,
 
   const MemoryAccess &Access = Statement.getAccessFor(Load);
 
+  // Make sure we have scalar values available to access the pointer to
+  // the data location.
+  extractScalarValues(Load, VectorMap, ScalarMaps);
+
   Value *NewLoad;
   if (Access.isStrideZero(isl_map_copy(Schedule)))
     NewLoad = generateStrideZeroLoad(Load, ScalarMaps[0]);
   else if (Access.isStrideOne(isl_map_copy(Schedule)))
-    NewLoad = generateStrideOneLoad(Load, ScalarMaps[0]);
+    NewLoad = generateStrideOneLoad(Load, ScalarMaps);
+  else if (Access.isStrideX(isl_map_copy(Schedule), -1))
+    NewLoad = generateStrideOneLoad(Load, ScalarMaps, true);
   else
     NewLoad = generateUnknownStrideLoad(Load, ScalarMaps);
 
@@ -556,6 +582,10 @@ void VectorBlockGenerator::copyStore(const StoreInst *Store,
   Value *Vector = getVectorValue(Store->getValueOperand(), VectorMap,
                                  ScalarMaps, getLoopForInst(Store));
 
+  // Make sure we have scalar values available to access the pointer to
+  // the data location.
+  extractScalarValues(Store, VectorMap, ScalarMaps);
+
   if (Access.isStrideOne(isl_map_copy(Schedule))) {
     Type *VectorPtrType = getVectorPtrTy(Pointer, VectorWidth);
     Value *NewPointer = getNewValue(Pointer, ScalarMaps[0], GlobalMaps[0],
@@ -579,10 +609,8 @@ void VectorBlockGenerator::copyStore(const StoreInst *Store,
 
 bool VectorBlockGenerator::hasVectorOperands(const Instruction *Inst,
                                              ValueMapT &VectorMap) {
-  for (Instruction::const_op_iterator OI = Inst->op_begin(),
-                                      OE = Inst->op_end();
-       OI != OE; ++OI)
-    if (VectorMap.count(*OI))
+  for (Value *Operand : Inst->operands())
+    if (VectorMap.count(Operand))
       return true;
   return false;
 }
@@ -593,10 +621,8 @@ bool VectorBlockGenerator::extractScalarValues(const Instruction *Inst,
   bool HasVectorOperand = false;
   int VectorWidth = getVectorWidth();
 
-  for (Instruction::const_op_iterator OI = Inst->op_begin(),
-                                      OE = Inst->op_end();
-       OI != OE; ++OI) {
-    ValueMapT::iterator VecOp = VectorMap.find(*OI);
+  for (Value *Operand : Inst->operands()) {
+    ValueMapT::iterator VecOp = VectorMap.find(Operand);
 
     if (VecOp == VectorMap.end())
       continue;
@@ -610,10 +636,11 @@ bool VectorBlockGenerator::extractScalarValues(const Instruction *Inst,
       // If there is one scalar extracted, all scalar elements should have
       // already been extracted by the code here. So no need to check for the
       // existance of all of them.
-      if (SM.count(*OI))
+      if (SM.count(Operand))
         break;
 
-      SM[*OI] = Builder.CreateExtractElement(NewVector, Builder.getInt32(i));
+      SM[Operand] =
+          Builder.CreateExtractElement(NewVector, Builder.getInt32(i));
     }
   }
 
@@ -712,7 +739,6 @@ void VectorBlockGenerator::copyBB() {
   VectorValueMapT ScalarBlockMap(getVectorWidth());
   ValueMapT VectorBlockMap;
 
-  for (BasicBlock::const_iterator II = BB->begin(), IE = BB->end(); II != IE;
-       ++II)
-    copyInstruction(II, VectorBlockMap, ScalarBlockMap);
+  for (Instruction &Inst : *BB)
+    copyInstruction(&Inst, VectorBlockMap, ScalarBlockMap);
 }
